from fastapi import Query
from fastapi_pagination import LimitOffsetPage, LimitOffsetParams, set_page
from fastapi_pagination.api import resolve_params
from fastapi_pagination.ext.sqlalchemy import paginate


set_page(LimitOffsetPage)


class PaginationParams(LimitOffsetParams):
    """Here we can redefine default size value"""

    limit: int = Query(1000, ge=1, le=1000, description="Page size")


def apply_pagination(query, db, params, response):
    """Apply pagination for the query
    * Paginate query according to applied parameters (size and page)
    * Add pagination headers to the response
    Don't filter or remove elements from query after the pagination
    """
    paginated = paginate(db, query, params)
    resolve_params(params)
    init = paginated.offset
    # If we are selecting a page that it's out of range, we return the same start and end for on
    # header range, copying the same behavior as connect.
    end = init
    if paginated.items:
        end += len(paginated.items) - 1

    response.headers['Content-Range'] = f'items {init}-{end}/{paginated.total}'
    return paginated.items



from typing import Any, Dict, Union

from fastapi import status

from connect.client import ClientError


'''
Error utilities for EaaS extensions.

By default eass runner injects the `connect.eaas.core.utils.client_error_exception_handler`
function to the FastAPI app initialization as exception handler.
This function catch all errors raised in the context of the WebApplication that are
an instance of the `connecet.client.ClientError` class and coerce them into JSON format
(`fastapi.responses.JSONResponse`) if the error instance has the attribute `error_code` setted.

`connect_extension_utils.api.errors` is a wrapper around this behaviour that facilitates the use of
prefixed error responses.
'''


class Error:
    STATUS_CODE = status.HTTP_400_BAD_REQUEST

    def __init__(self, message, error_code):
        self.message = message
        self.error_code = error_code

    def __call__(self, **kwds: Dict[str, Any]) -> ClientError:
        format_kwargs = kwds.get('format_kwargs', {})

        message = self.message.format(**format_kwargs)
        errors = kwds.get('errors')

        if not errors:
            errors = [message or 'Unexpected error.']
        if not isinstance(errors, list):
            errors = [errors]

        return ClientError(
            message=message,
            status_code=self.STATUS_CODE,
            error_code=self.error_code,
            errors=errors,
        )


class ExtensionErrorMeta(type):
    PREFIX = 'EXT'
    ERRORS = {}

    def __getattr__(cls, __name: str) -> Union[Error, AttributeError]:
        valid_dict = {cls.PREFIX: cls.ERRORS}
        try:
            prefix, code = __name.split('_')
            error = valid_dict[prefix][int(code)]
        except (KeyError, ValueError):
            raise AttributeError(f"type object '{cls.__name__}' has no attribute '{__name}'")
        return Error(message=error, error_code=__name)


class ExtensionErrorBase(metaclass=ExtensionErrorMeta):
    '''
    Base Error class to group a set of validation (`fastapi.status.HTTP_400_BAD_REQUEST`)
    errors base on a prefix. By default the `PREFIX` value is `EXT`, but it can be overwritten.
    Also a list of `errors` can be provided.

    Usage:

    ```
    # Define a custom error class
    class MyError(ExtensionErrorBase)
        PREFIX = "EXT"
        ERRORS = {
            1: "Some error",
            2: "Some {template} error.",
            3: "Not found",

        }

    # raise the error
    raise MyError.EXT_001()
    raise MyError.EXT_002(format_kwargs={"template": "foo"})
    ```
    '''


class Http404(ClientError):
    def __init__(self, obj_id, **kwargs):
        message = "Object `{obj_id}` not found.".format(obj_id=obj_id)
        status_code = status.HTTP_404_NOT_FOUND
        error_code = 'NFND_000'
        errors = [message]
        super().__init__(message, status_code, error_code, errors, **kwargs)



import os
import random
from contextlib import contextmanager

from fastapi import Depends
from sqlalchemy import create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.exc import DontWrapMixin
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session, sessionmaker

from connect.eaas.core.inject.common import get_config


_MAX_RETRIES = 1000
_ENGINE = None


def _get_numeric_string(size):
    sys_random = random.SystemRandom()
    return str(sys_random.randint(1 * 10 ** (size - 1), 1 * 10**size - 1))


def _generate_verbose_id(prefix):
    return (
        f"{prefix}-{_get_numeric_string(3)}" f"-{_get_numeric_string(3)}-{_get_numeric_string(3)}"
    )


class VerboseSessionError(Exception, DontWrapMixin):
    """
    Exception class to handle errors through custom methods of `VerboseBaseSession`.
    """


class VerboseBaseSession(Session):
    """
    Custom `sqlalchemy.orm.Session` class to generate and add verbose_id to a model.
    """

    def _set_verbose(self, instance):
        instance_class = instance.__class__
        for _ in range(1, _MAX_RETRIES + 1):
            verbose_id = _generate_verbose_id(instance_class.PREFIX)
            if not (
                self.query(
                    self.query(instance_class).filter(instance_class.id == verbose_id).exists(),
                ).scalar()
            ):
                instance.id = verbose_id
                return instance

        raise VerboseSessionError(
            f"Could not generate {instance_class.__name__} verbose ID"
            f" after {_MAX_RETRIES} attempts.",
        )

    def _set_verbose_all(self, instances):
        checked_instances = []
        if instances:
            count = len(instances)
            instance_class = instances[0].__class__
            all_same_class_check = all(isinstance(ins, instance_class) for ins in instances)
            assert all_same_class_check, "All instances must be of the same class."
            instance_class = instances[0].__class__
            for _ in range(1, _MAX_RETRIES + 1):
                ids = [_generate_verbose_id(instance_class.PREFIX) for _ in range(count)]
                if not (
                    self.query(
                        self.query(instance_class).filter(instance_class.id.in_(ids)).exists(),
                    ).scalar()
                ):
                    break
                ids = []
            if not ids:
                raise VerboseSessionError(
                    f"Could not generate a group of {count} {instance_class.__name__} verbose ID"
                    f" after {_MAX_RETRIES} attempts.",
                )

            for instance, verbose_id in zip(instances, ids):
                instance.id = verbose_id
                checked_instances.append(instance)
        return checked_instances

    def add_with_verbose(self, instance):
        instance = self._set_verbose(instance)
        return self.add(instance)

    def add_all_with_verbose(self, instances):
        instances = self._set_verbose_all(instances)
        return self.add_all(instances)

    def add_next_with_verbose(self, instance, related_id_field):
        instance_class = instance.__class__
        new_suffix = 0
        related_id_value = getattr(instance, related_id_field)
        last_obj = self._get_last_obj_for_next_verbose(
            instance_class,
            related_id_field,
            related_id_value,
        )
        if last_obj:
            _instance_id, suffix = last_obj.id.rsplit("-", 1)
            new_suffix = int(suffix) + 1
        else:
            id_body = related_id_value.split("-", 1)[-1]
            _instance_id = f"{instance_class.PREFIX}-{id_body}"

        instance.id = "{0}-{1}".format(_instance_id, "{0:03d}".format(new_suffix))
        return self.add(instance)

    def add_all_with_next_verbose(self, instances, related_id_field):
        first_item = instances[0]
        instance_class = first_item.__class__
        new_suffix = 0
        related_id_value = getattr(first_item, related_id_field)
        last_obj = self._get_last_obj_for_next_verbose(
            instance_class,
            related_id_field,
            related_id_value,
        )
        if last_obj:
            _instance_id, suffix = last_obj.id.rsplit("-", 1)
            new_suffix = int(suffix) + 1
        else:
            id_body = related_id_value.split("-", 1)[-1]
            _instance_id = f"{instance_class.PREFIX}-{id_body}"

        for instance in instances:
            instance.id = "{0}-{1}".format(_instance_id, "{0:03d}".format(new_suffix))
            new_suffix += 1

        return self.add_all(instances)

    def _get_last_obj_for_next_verbose(self, model_class, related_id_field, related_id_value):
        base_qs = self.query(model_class).filter(
            model_class.__dict__[related_id_field] == related_id_value,
        )
        last_obj = None
        if self.query(base_qs.exists()).scalar():
            last_obj = base_qs.order_by(
                model_class.id.desc(),
            ).first()
        return last_obj


SessionLocal = sessionmaker(autocommit=False, autoflush=False, class_=VerboseBaseSession)
Model = declarative_base()


def get_engine(config: dict = Depends(get_config)):
    global _ENGINE
    if not _ENGINE:
        _ENGINE = create_engine(
            config.get("DATABASE_URL", os.getenv("DATABASE_URL")),
            pool_pre_ping=True,
            pool_recycle=300,
        )

    return _ENGINE


def get_db(engine: Engine = Depends(get_engine)):
    db: VerboseBaseSession = SessionLocal(bind=engine)
    try:
        yield db
    finally:
        db.close()


def create_db(config: dict = Depends(get_config)):
    engine = get_engine(config)
    Model.metadata.create_all(bind=engine)
    return engine


@contextmanager
def get_db_ctx_manager(config):
    engine: Engine = get_engine(config)
    db: VerboseBaseSession = SessionLocal(bind=engine)
    try:
        yield db
    finally:
        db.close()



from datetime import datetime
from typing import Dict, Optional, Union

from pydantic import BaseModel, root_validator


_By = Optional[Union[str, Dict[str, str]]]
Events = Dict[str, Dict[str, Union[datetime, _By]]]


def clean_empties_from_dict(data):
    """
    Removes inplace all the fields that are None or empty dicts in data.
    Returns param data, that was modified inplace.
    If the param is not a dict, will return the param unmodified.
    :param data: dict
    :rtype: dict
    """
    if not isinstance(data, dict):
        return data

    for key in list(data.keys()):
        value = data[key]
        if isinstance(value, dict):
            clean_empties_from_dict(value)
            value = data[key]
        if not value:
            del data[key]
    return data


class NonNullSchema(BaseModel):
    """
    Custom `pydantic.BaseModel` class that exlude Nullable fields if they are `None` at
    runtime resulting in an object representation without null
    values.
    """

    def dict(self, *args, **kwargs):
        kwargs['exclude_none'] = True
        return super().dict(*args, **kwargs)

    @root_validator(pre=True)
    def validate_events(cls, values):
        events = values.get('events')
        if events:
            values['events'] = clean_empties_from_dict(events)
        return values


class ReferenceSchema(NonNullSchema):
    """
    Helper Schema class for related object reference representation.
    """

    id: str
    name: Optional[str]
